import os 
import json
import re
import fitz  # PyMuPDF
import mimetypes
import warnings
from flask import Flask, request, render_template, jsonify
from passporteye import read_mrz
import pytesseract
from PIL import Image
from datetime import datetime, timedelta
import cv2
import matplotlib.image as mpimg
import easyocr

warnings.filterwarnings('ignore')

app = Flask(__name__)
app.config['UPLOAD_FOLDER'] = 'uploads'
reader = easyocr.Reader(lang_list=['en'], gpu=False)

# Load country codes
with open('country_codes.json') as f:
    country_codes = json.load(f)

# Helper Functions

def parse_date(string):
    """Parse date string from MRZ format (YYMMDD) to DD/MM/YYYY."""
    try:
        return datetime.strptime(string, '%y%m%d').strftime('%d/%m/%Y')
    except ValueError:
        return None

def get_country_name(country_code):
    """Retrieve full country name from the 3-letter country code."""
    return next((country['name'].upper() for country in country_codes if country['alpha-3'] == country_code), country_code)

def get_sex(code):
    """Determine gender based on the MRZ code."""
    code = code.upper()
    return 'MALE' if code in ['M', 'MALE', '0'] else 'FEMALE' if code in ['F', 'FEMALE'] else 'UNKNOWN'

def extract_mrz_data(img_path):
    """Extract MRZ data from passport image with improved name handling."""
    img_name = img_path
    mrz = read_mrz(img_name, save_roi=True)
    if not mrz:
        return {"error": "MRZ not found or unreadable"}

    temp_image_path = 'tmp_mrz.png'
    data = mrz.to_dict()
    mpimg.imsave(temp_image_path, mrz.aux['roi'], cmap='gray')
    img = cv2.imread(temp_image_path)

    # Parse MRZ data
    passport_number = data.get("number", "").replace("<", "").strip()
    issuing_country = get_country_name(data.get("country"))
    nationality = get_country_name(data.get("nationality"))
    date_of_birth = parse_date(data.get("date_of_birth"))
    expiration_date = parse_date(data.get("expiration_date"))
    sex = get_sex(data.get("sex"))

    img = cv2.resize(img, (1110, 140))
    mrz_lines = reader.readtext(img, paragraph=False, detail=0)
    
    if len(mrz_lines) < 2:
        return {'error': "Failed to retrieve MRZ lines from the image"}
    a, b = mrz_lines[0].upper(), mrz_lines[1].upper()
    a = (a + '<' * 44)[:44]
    b = (b + '<' * 44)[:44]
    
    surname_names = a[5:44].split('<<', 1)
    surname, names = (surname_names + [''])[:2]

    return {
        "passport_number": passport_number,
        "issuing_country": issuing_country,
        "surname": surname.replace('<', ' ').strip(),
        "name": names.replace('<', ' ').strip(),
        "sex": sex,
        "date_of_birth": date_of_birth,
        "expiration_date": expiration_date,
        "nationality": nationality,
    }

def extract_issuance_date(img_path):
    """Extract issuance date from passport using OCR."""
    text = pytesseract.image_to_string(Image.open(img_path))
    match = re.search(r"(Date of Issue|Issued On)\s*[:\s]\s*(\d{2}/\d{2}/\d{4})", text, re.IGNORECASE)
    return match.group(2) if match else "Not found"

def extract_text_from_pdf(pdf_path):
    """Extract text from a PDF file using PyMuPDF; falls back to OCR if no text is found."""
    text = ""
    with fitz.open(pdf_path) as pdf:
        for page_num in range(pdf.page_count):
            page = pdf[page_num]
            page_text = page.get_text()
            if not page_text.strip():
                pix = page.get_pixmap()
                img = Image.frombytes("RGB", [pix.width, pix.height], pix.samples)
                page_text = pytesseract.image_to_string(img)
            text += page_text + "\n"
    return text

def find_dates_from_text(text):
    """Extract dates based on specific patterns in the text."""
    today = datetime.today()
    results = {}

    patterns = [
        # 1. Date of Expiration:01.11.2024
        (r"Date of Expiration:\s*(\d{2}\.\d{2}\.\d{4})", "Expiration Date"),

        # 2. Deadline for Deposit Payment: 07.11.2024
        (r"Deadline for Deposit Payment:\s*(\d{2}\.\d{2}\.\d{4})", "Deposit Payment Deadline"),

        # 3. the bank receipt must be uploaded to our online system in 15 days
        (r"the bank receipt must be uploaded to our\s*\n\s*online system in (\d+) days", 
         lambda days: today + timedelta(days=int(days)), "Bank Receipt Deadline"),

        # 4. Payment Deadline 07 November 2024 (date in new line)
        (r"Payment Deadline\s*\n(\d{1,2}\s\w+\s\d{4})", "Payment Deadline"),

        # 5. Offer letter valid for 10 days from today
        (r"This offer letter is valid for (\d+) days starting from", 
         lambda days: today + timedelta(days=int(days)), "Offer Letter Validity Deadline"),

        # 6. Payment until 01/11/2024, date on the same line
        (r"complete your payment until (\d{2}/\d{2}/\d{4})", "Payment Due Date"),

        # 7. Wire transfer within 3 days from today
        (r"wire transfer must be completed in (\d+) days", 
         lambda days: today + timedelta(days=int(days)), "Wire Transfer Deadline"),

        # 8. Starting from 30.10.2024 within 10 days, date in new line
        (r"starting\s*\nfrom (\d{2}\.\d{2}\.\d{4}) .* within (\d+) days", 
         lambda start_date, days: datetime.strptime(start_date, "%d.%m.%Y") + timedelta(days=int(days)), 
         "Acceptance Deadline"),

        # 9. Date of Expiration with spaces: 01.11.2024
        (r"Date of Expiration:\s*(\d{2}\.\d{2}\.\d{4})", "Expiration Date (Alternative)"),

        # 10. Offer letter valid for 10 days from today
        (r"This offer letter is valid for (\d+) days starting from", 
         lambda days: today + timedelta(days=int(days)), "Offer Letter Validity Deadline (Alternative)"),

        # 11. Deadline for Registration Payment 30/10/2024
        (r"Deadline for Registration Payment\s*(\d{2}/\d{2}/\d{4})", "Registration Payment Deadline"),

        # 12. Prepayment within 15 days from today
        (r"prepayment within (\d+) days of receiving", 
         lambda days: today + timedelta(days=int(days)), "Prepayment Deadline"),

        # 13. Deposit fee not paid until 09/11/2024
        (r"deposit fee is not paid until (\d{2}/\d{2}/\d{4})", "Deposit Fee Deadline"),

        # 14. Turkish payment deadline date only 28.10.2024
        (r"tarihinden itibaren\s*(\d{2}\.\d{2}\.\d{4})\s*\(dahil\)", "Turkish Payment Deadline"),

        # 15. Acceptance within 20 days, starting from 25/10/2024, date on new line
        (r"complete your acceptance\s*\nwithin (\d+) days, starting from (\d{2}/\d{2}/\d{4})", 
         lambda days, start_date: datetime.strptime(start_date, "%d/%m/%Y") + timedelta(days=int(days)), 
         "Acceptance Completion Deadline"),

        # 16. Valid until the end of 28.10.2024
        (r"valid until the end of (\d{2}\.\d{2}\.\d{4})", "Conditional Acceptance Validity End Date"),

        # 17. Deposit Deadline with date on a new line
        (r"Deposit Deadline\s*\n\s*:\s*(\d{1,2}\s\w+\s\d{4})", "Deposit Deadline"),

        # 18. Proof of payment within 7 days from today
        (r"proof of payment.*within (\d+) days", 
         lambda days: today + timedelta(days=int(days)), "Proof of Payment Deadline"),

        # 19. Registration Deadline*: 31.10.2024
        (r"Registration Deadline\*\s*:\s*(\d{2}\.\d{2}\.\d{4})", "Registration Deadline"),

        # 20. Payment until 2024-10-28 in ISO format
        (r"deposit payment until (\d{4}-\d{2}-\d{2})", "Deposit Payment Deadline (ISO)"),

        # 21. Failure to pay within ONE week
        (r"Failure to pay within ONE week", 
         lambda _: today + timedelta(days=7), "One Week Payment Deadline"),

        # 22. Payment before 21.10.2024
        (r"pay deposit before (\d{2}\.\d{2}\.\d{4})", "Deposit Payment Cutoff Date"),

        # 23. Enrollment deposit payment within 10 days from today
        (r"enrollment deposit payment within (\d+) days", 
         lambda days: today + timedelta(days=int(days)), "Enrollment Deposit Deadline"),
        (r"enrollment depoisit payment within (\d+) days", 
         lambda days: today + timedelta(days=int(days)), "Enrollment Deposit Deadline"),
        
        # 24. Due date for advanced tuition payment 21.10.2024
        (r"due date for advanced tuition payment.*(\d{2}\.\d{2}\.\d{4})", "Advanced Tuition Payment Due Date"),

        # 25. Conditional acceptance valid for 14 days from today
        (r"conditional acceptance.*valid for (\d+) days starting from", 
         lambda days: today + timedelta(days=int(days)), "Conditional Acceptance Validity"),

        # 26. Deposit fee until 17/09/2024
        (r"deposit fee.*until (\d{2}/\d{2}/\d{4})", "Deposit Fee Deadline (Specific)"),

        # 27. Pre-Payment Duration 10 days from today
        (r"Pre-Payment Duration\s*(\d+) days", 
         lambda days: today + timedelta(days=int(days)), "Pre-Payment Duration Deadline"),

        # 28. Acceptance within 10 days, starting from 18.09.2024
        (r"complete your acceptance within (\d+) days", 
         lambda days: today + timedelta(days=int(days)), "Acceptance Completion Deadline from Today"),

        # 29. Finalize payment within 10 days from today
        (r"finalize your payment transactions within (\d+) days", 
         lambda days: today + timedelta(days=int(days)), "Final Payment Transaction Deadline"),

        # 30. Wire transfer within 15 days from today
        (r"wire transfer.*within (\d+) days", 
         lambda days: today + timedelta(days=int(days)), "Wire Transfer Completion Deadline"),

        # 31. Deposit Payment Deadline:21.10.2024
        (r"Deposit Payment Deadline:(\d{2}\.\d{2}\.\d{4})", "Deposit Payment Deadline Alternative"),

        # 32. Conditional acceptance valid for 7 days from today
        (r"conditional acceptance.*valid for (\d+)(?:\s*\(\w+\))?\s+days", 
         lambda days: today + timedelta(days=int(days)), "Conditional Acceptance Validity (Alternative)"),

        # 33. Deadline for Registration Payment 23/10/2024
        (r"Deadline for Registration Payment\s*(\d{2}/\d{2}/\d{4})", "Alternative Registration Payment Deadline"),

        # 34. Receipt submission before 01/10/2024
        (r"receipt.*before (\d{2}/\d{2}/\d{4})", "Receipt Submission Deadline"),
        
        (r"Conditional acceptance valid until\s*\n(\d{2}\.\d{2}\.\d{4})", "Conditional Acceptance Expiration Date"),
        (r"conditional acceptance.*valid only for (\d+) days starting from", 
         lambda days: today + timedelta(days=int(days)), "Conditional Acceptance Validity from Today"),

        
        (r"bank account within (\d+) days starting from today", 
         lambda days: today + timedelta(days=int(days)), 
         "Bank Account Deadline"),
    ]

    
    for pattern, *args in patterns:
        if callable(args[0]):
            matches = re.findall(pattern, text)
            for match in matches:
                # Handle callable-based dates
                calculated_date = args[0](match) if isinstance(match, str) else args[0](*match)
                results[args[1]] = calculated_date.strftime("%d-%m-%Y")
        else:
            match = re.search(pattern, text)
            if match:
                raw_date = match.group(1)
                date_obj = None

                # Try parsing based on known formats
                try:
                    if "." in raw_date:
                        date_obj = datetime.strptime(raw_date, "%d.%m.%Y")
                    elif "/" in raw_date:
                        date_obj = datetime.strptime(raw_date, "%d/%m/%Y")
                    elif "-" in raw_date:
                        date_obj = datetime.strptime(raw_date, "%Y-%m-%d")
                    else:
                        date_obj = datetime.strptime(raw_date, "%d %B %Y")
                except ValueError:
                    continue  # Skip if the date format is unrecognized

                # Add to results with uniform formatting
                if date_obj:
                    results[args[0]] = date_obj.strftime("%d-%m-%Y")

    return results

# Flask Routes
@app.route('/')
def index():
    return render_template('index.html')

@app.route('/upload/passport', methods=['GET', 'POST'])
def upload_passport():
    if request.method == 'POST':
        file = request.files.get('file')
        if not file:
            return render_template('upload_passport.html', error="No file uploaded")

        file_path = os.path.join(app.config['UPLOAD_FOLDER'], file.filename)
        file.save(file_path)

        mrz_data = extract_mrz_data(file_path)
        issuance_date = extract_issuance_date(file_path)
        mrz_data["issuance_date"] = issuance_date

        os.remove(file_path)
        return render_template('upload_passport.html', data=mrz_data)
    
    return render_template('upload_passport.html')

@app.route('/upload/conditional-acceptance', methods=['GET', 'POST'])
def upload_conditional_acceptance():
    if request.method == 'POST':
        file = request.files.get('file')
        if not file:
            return render_template('upload_conditional_acceptance.html', error="No file uploaded")

        file_path = os.path.join(app.config['UPLOAD_FOLDER'], file.filename)
        file.save(file_path)

        text = extract_text_from_pdf(file_path)
        date_results = find_dates_from_text(text)
        os.remove(file_path)

        if date_results:
            first_label, first_date = next(iter(date_results.items()))
            return render_template('upload_conditional_acceptance.html', admission_deadline=first_date)
        else:
            return render_template('upload_conditional_acceptance.html', error="No relevant dates found")
    
    return render_template('upload_conditional_acceptance.html')

@app.route('/api/upload/passport', methods=['POST'])
def api_upload_passport():
    file = request.files.get('file')
    if not file:
        return jsonify({"error": "No file uploaded"}), 400

    file_path = os.path.join(app.config['UPLOAD_FOLDER'], file.filename)
    file.save(file_path)

    mime_type, _ = mimetypes.guess_type(file_path)
    if mime_type not in ['image/jpeg', 'image/png']:
        os.remove(file_path)
        return jsonify({"error": "Unsupported file type, please upload a JPEG or PNG image"}), 400

    mrz_data = extract_mrz_data(file_path)
    issuance_date = extract_issuance_date(file_path)
    mrz_data["issuance_date"] = issuance_date
    os.remove(file_path)
    
    return jsonify(mrz_data)

@app.route('/api/upload/pdf', methods=['POST'])
def api_upload_pdf():
    file = request.files.get('file')
    if not file:
        return jsonify({"error": "No file uploaded"}), 400

    file_path = os.path.join(app.config['UPLOAD_FOLDER'], file.filename)
    file.save(file_path)

    text = extract_text_from_pdf(file_path)
    date_results = find_dates_from_text(text)
    os.remove(file_path)

    if date_results:
        first_label, first_date = next(iter(date_results.items()))
        return jsonify({"admission_deadline": first_date})
    else:
        return jsonify({"admission_deadline": "Not found"}), 404

if __name__ == '__main__':
    os.makedirs(app.config['UPLOAD_FOLDER'], exist_ok=True)
    app.run(host='0.0.0.0', port=5000, debug=True)
